/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: flow-receive.c,v 1.28 2001/06/22 19:52:53 maf Exp $
 */

#if HAVE_CONFIG_H
 #include <config.h>
#endif


#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_STRINGS_H
 #include <strings.h>
#endif
#if HAVE_STRING_H
  #include <string.h>
#endif
#include <time.h>
#include <fcntl.h>
#include <support.h>
#include <ftlib.h>
#include <fmt.h>

int debug;
int done;

void usage();
void sig_quit();

int main(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  extern int optind, errno;
  extern int done;
#ifdef IP_ADD_MEMBERSHIP
  struct sockaddr_in tmp_addr;
  struct ip_mreq mr;
#ifdef IP_ADD_SOURCE_MEMBERSHIP
  struct ip_mreq_source mrs;
#endif
#endif
  struct ftio ftio;
  struct ftset ftset;
  struct ftpdu ftpdu;
  struct ftexporter ftx;
  struct ftpeeri ftpi;
  struct ftipmask ftipmask;
  struct stat stat_buf;
  int len, i, n, offset, out_fd, out_fd_plain;
  unsigned int v1, v2;
  fd_set rfd;
  char fmt_buf[32], *out_fname;
  u_int32 nflows, time_start, time_end;
  u_int32 flows_corrupt, flows_lost, flows_reset;
  u_int32 privacy_mask;

  privacy_mask = 0xFFFFFFFF;
  out_fname = (char*)0L;
  nflows = 0;
  out_fd_plain = 0;
  out_fd = -1;
  ftset.z_level = 0;
  done = 0;
  bzero(&ftx, sizeof ftx);
  bzero(&ftpdu, sizeof ftpdu);
  flows_corrupt = flows_lost = flows_reset = 0;

  /* init fterr */
  fterr_setid(argv[0]);

  /* defaults + no compression */
  ftset_init(&ftset, 0);

  /* listen for PDU's */
  ftx.cli_addr.sin_family = AF_INET;
  ftx.cli_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  ftx.cli_addr.sin_port = htons(FT_PORT);

  while ((i = getopt(argc, argv, "A:b:C:d:h?m:o:V:z:")) != -1)

    switch (i) {

    case 'A': /* AS substitution */
      ftset.as_sub = atoi(optarg);
      break;

    case 'b': /* output byte order */
      if (!strcasecmp(optarg, "little"))
        ftset.byte_order = FT_HEADER_LITTLE_ENDIAN;
      else if (!strcasecmp(optarg, "big"))
        ftset.byte_order = FT_HEADER_BIG_ENDIAN;
      else 
        fterr_errx(1, "expecting \"big\" or \"little\"");
      break;

    case 'C': /* comment field */
      ftset.comments = optarg;
      break;

    case 'd': /* debug */
      debug = atoi(optarg);
      break;

    case 'h': /* help */
    case '?':
      usage();
      exit (0);
      break;

    case 'm': /* privacy mask */ 
      privacy_mask = scan_ip(optarg);
      break;

   case 'o': /* output filename */
      out_fname = optarg;
      break;

   case 'V': /* PDU version */
      n = sscanf(optarg, "%u.%u", &v1, &v2);
      if (n == 1) {
        ftx.ftv.s_version = FT_IO_SVERSION;
        ftx.ftv.d_version = v1;
        ftx.ftv.set = 1;
      } else if (n == 2) {
        ftx.ftv.s_version = FT_IO_SVERSION;
        ftx.ftv.d_version = v1;
        ftx.ftv.agg_method = v2;
        ftx.ftv.agg_version = 2;
        ftx.ftv.set = 1;
      } else
        fterr_errx(1, "Version scan failed");
      break;


    case 'z': /* compress level */
      ftset.z_level = atoi(optarg);
      if ((ftset.z_level < 0) || (ftset.z_level > 9))
        fterr_errx(1, "Compression level must be between 0 and 9");
      break;

    default:
      usage();
      exit (1);
      break;

    } /* switch */

  /* setup for ftrec_mask_ip */
  if (privacy_mask != 0xFFFFFFFF)
    ftrec_compute_mask(&ftipmask, privacy_mask, privacy_mask,
      ftset.byte_order);

  /* number of unprocessed options */
  n = argc - optind;

  /* locip/remip/port */
  if (n == 1) {

    ftpi = scan_peeri(argv[optind]);

    ftx.ip = ftpi.remip;

    if (ftpi.dstport)
      ftx.port = ftpi.dstport;
    else
      ftx.port = FT_PORT;

    ftx.cli_addr.sin_addr.s_addr = htonl(ftpi.locip);
    ftx.cli_addr.sin_port = htons(ftx.port);

  } else if (n == 0) {
    /* defaults */
  } else {
    fterr_errx(1, "Illegal option: %s", argv[optind+1]);
  }

  /* IP address of exporter should be set */
  if ((IN_CLASSD(ftpi.remip) && (!ftpi.locip)) ||
      ((!IN_CLASSD(ftpi.remip)) && (!ftpi.remip)))
    fterr_warnx("Exporter IP not specified");

  /* handle SIGQUIT */
  signal(SIGQUIT, sig_quit);
  signal(SIGHUP, sig_quit);
  signal(SIGINT, sig_quit);
  signal(SIGTERM, sig_quit);

  /* get hostname */
  if (gethostname((char*)&ftset.hnbuf, (int)FT_HOSTNAME_LEN-1) == -1)
    fterr_err(1, "gethostname()");

  ftset.hnbuf[FT_HOSTNAME_LEN-1] = 0;

  if ((ftx.fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    fterr_err(1, "socket()");

  if (bigsockbuf(ftx.fd, SO_RCVBUF, 224*1024) < 0)
    fterr_err(1, "bigsockbuf()");

/* multicast capable? */
#ifdef IP_ADD_MEMBERSHIP
  
  if (IN_CLASSD(ftpi.remip)) {

    /* source is the first arg now */
    ftx.ip = ftpi.locip;

    /* socket API usually requires INADDR_ANY
     * and s/g/port identifier does not have a source interface field
     * to use here
     */
    bzero(&tmp_addr, sizeof tmp_addr);
    tmp_addr.sin_family = AF_INET;
    tmp_addr.sin_port = htons(ftx.port);

    if (bind(ftx.fd, (struct sockaddr*)&tmp_addr,
      sizeof(struct sockaddr)) < 0)
      fterr_err(1, "bind(mcast-rcv)");

#ifdef IP_ADD_SOURCE_MEMBERSHIP

    /* ssm address? */
    if (IN_CLASSD_SSM(ftpi.remip)) {

      mrs.imr_sourceaddr.s_addr = htonl(ftpi.locip);
      mrs.imr_multiaddr.s_addr = htonl(ftpi.remip);
      mrs.imr_interface.s_addr = INADDR_ANY;

      if (setsockopt(ftx.fd, IPPROTO_IP, IP_ADD_SOURCE_MEMBERSHIP,
        (char*)&mrs, sizeof(mrs)) < 0)
        fterr_err(1, "setsockopt(IP_ADD_SOURCE_MEMBERSHIP)");

    }

    goto mcast_done;
#endif /* IP_ADD_SOURCE_MEMBERSHIP */

    mr.imr_multiaddr.s_addr = htonl(ftpi.remip);
    mr.imr_interface.s_addr = INADDR_ANY;
    
    if (setsockopt(ftx.fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
      (char *)&mr, sizeof(mr)) < 0)
      fterr_err(1, "setsockopt(IP_ADD_MEMBERSHIP)");


  } else { /* is a multicast group */

    /* unicast bind -- multicast support */
    if (bind(ftx.fd, (struct sockaddr*)&ftx.cli_addr, sizeof(ftx.cli_addr)) < 0)
      fterr_err(1, "bind()");

  } /* not multicast group */

#ifdef IP_ADD_SOURCE_MEMBERSHIP
mcast_done:
#endif /* IP_ADD_SOURCE_MEMBERSHIP */

#else /* IP_ADD_MEMBERSHIP */

  /* unicast bind -- no multicast support */
  if (bind(ftx.fd, (struct sockaddr*)&ftx.cli_addr, sizeof(ftx.cli_addr)) < 0)
    fterr_err(1, "bind()");

#endif /* IP_ADD_MEMBERSHIP */

  len = sizeof(ftx.srv_addr);

  /* if out_fname is not set, then use stdout */
  if (out_fname) {

    if ((out_fd = open(out_fname,  O_WRONLY|O_CREAT|O_TRUNC, 0644)) == -1)
      fterr_err(1, "open(%s)", out_fname);

    if (fstat(out_fd, &stat_buf) == -1)
      fterr_err(1, "fstat(%s)", out_fname);
      
    /* is this a plain file? */
    if (!stat_buf.st_rdev)
      out_fd_plain = 1;
    
  } else
    out_fd = 1;


  /* output to out_fd */
  if (ftio_init(&ftio, out_fd, FT_IO_FLAG_NO_SWAP | FT_IO_FLAG_WRITE |
    ((ftset.z_level) ? FT_IO_FLAG_ZINIT : 0) ) < 0)
    fterr_errx(1, "ftio_init(): failed");

  time_start = (u_int32)time((time_t)0L);

  ftio_set_comment(&ftio, ftset.comments);
  ftio_set_cap_hostname(&ftio, ftset.hnbuf);
  ftio_set_byte_order(&ftio, ftset.byte_order);
  ftio_set_z_level(&ftio, ftset.z_level);
  ftio_set_cap_time(&ftio, time_start, 0);
  ftio_set_debug(&ftio, debug);
  ftio_set_streaming(&ftio, 1);

  ftio_map_load(&ftio, FT_MAP_FILE, ftx.ip);

  /* header must be full size on initial write */
  if (out_fd_plain) {
    ftio_set_flows_count(&ftio, nflows);
    ftio_set_corrupt(&ftio, flows_corrupt);
    ftio_set_lost(&ftio, flows_lost);
    ftio_set_reset(&ftio, flows_reset);
    ftio_set_xip(&ftio, ftx.ip);
    ftio_set_flows_count(&ftio, nflows);
  }

  /* if version set on command line, write out header here */
  if (ftx.ftv.set) {

    if (ftio_set_ver(&ftio, &ftx.ftv) < 0)
      fterr_errx(1, "ftio_set_ver(): failed");

    /* header first */
    if (ftio_write_header(&ftio) < 0)
      fterr_errx(1, "ftio_write_header()");

  }

  while (1) {

    FD_ZERO (&rfd);
    FD_SET (ftx.fd, &rfd);

    if (select (ftx.fd+1, &rfd, (fd_set *)0, (fd_set *)0, 
      (struct timeval*)0L) < 0)  {
      if (errno != EINTR)
        fterr_err(1, "select()");
    }

    if (done) {
      fterr_info("Cleaning up");
      break;
    }

    if (FD_ISSET(ftx.fd, &rfd)) {

      if ((ftpdu.len = recvfrom(ftx.fd, (char*)&ftpdu.buf, sizeof ftpdu.buf,
        0, (struct sockaddr*)&ftx.srv_addr, &len)) == -1)
        fterr_err(1, "recvfrom()");

      if (ftx.ip && ftx.ip != ntohl(ftx.srv_addr.sin_addr.s_addr)) {
        fterr_warnx("src IP (%s) not configured.  Skipping PDU.",
          inet_ntoa(ftx.srv_addr.sin_addr));
        flows_corrupt ++;
        goto skip1;
      } else { /* autoset */
        ftx.ip = ntohl(ftx.srv_addr.sin_addr.s_addr);
      }

      /* verify integrity, get version */
      if (ftpdu_verify(&ftpdu) < 0) {
        fterr_warnx("ftpdu_verify(): failed.  Skipping PDU");
        flows_corrupt ++;
        goto skip1;
      }

      if (ftpdu_check_seq(&ftpdu) < 0) {
        fterr_warnx(
          "ftpdu_seq_check(): fail, expecting=%lu received=%lu lost=%lu\n",
          (u_long)ftpdu.seq_exp, (u_long)ftpdu.seq_rcv,
          (u_long)ftpdu.seq_lost);

        /* only count these if lost if "lost" is a reasonable number */
        if (ftpdu.seq_lost < FT_SEQ_RESET)
          flows_lost += ftpdu.seq_lost;
        else   
          flows_reset ++;
      }

      /* first time? */
      if (!ftx.ftv.set) {

        /* set the version information in the io stream */
        if (ftio_set_ver(&ftio, &ftpdu.ftv) < 0)
          fterr_errx(1, "ftio_set_ver(): failed");

        /* copy to compare next time */
        bcopy(&ftpdu.ftv, &ftx.ftv, sizeof ftx.ftv);

        /* flag struct as configured */
        ftx.ftv.set = 1;

        /* header first */
        if (ftio_write_header(&ftio) < 0)
          fterr_errx(1, "ftio_write_header(): failed");

      } else {

        /* ensure previous version == current decoded pdu version */
        if ((ftx.ftv.d_version != ftpdu.ftv.d_version) ||
            (ftx.ftv.agg_method != ftpdu.ftv.agg_method)) {

          fterr_warnx("Unexpected PDU: od=%d d=%d oagg=%d agg=%d",
            (int)ftx.ftv.d_version, (int)ftpdu.ftv.d_version,
            (int)ftx.ftv.agg_method, (int)ftpdu.ftv.agg_method);

          flows_corrupt ++;

          goto skip1;

        } /* PDU version ok */

      } /* else */

      /* decode */
      ftx.ftd.byte_order = ftset.byte_order;
      ftx.ftd.as_sub = ftset.as_sub;
      ftx.ftd.exporter_ip = ftx.ip;
      n = fts3rec_pdu_decode(&ftpdu, &ftx.ftd);

      /* write decoded flows */
      for (i = 0, offset = 0; i < n; ++i, offset += ftx.ftd.rec_size) {

        ++nflows;

        /* simple data privacy */ 
        if (privacy_mask != 0xFFFFFFFF)
          ftrec_mask_ip(ftx.ftd.buf+offset, &ftx.ftv, &ftipmask);

        if (ftio_write(&ftio, ftx.ftd.buf+offset) < 0)
          fterr_errx(1, "ftio_write(): failed");

      } /* for */

skip1:

    } /* if FD_ISSET */
  } /* while 1 */

  fmt_uint32(fmt_buf, nflows, FMT_JUST_LEFT);
  fterr_info("%s flows processed\n", fmt_buf);

  /* rewrite header with updated info */
  if (out_fd_plain) {
  
    time_end = (u_int32)time((time_t)0L);

    ftio_set_xip(&ftio, ftx.ip);
    ftio_set_cap_time(&ftio, time_start, time_end);
    ftio_set_flows_count(&ftio, nflows);
    ftio_set_streaming(&ftio, 0);
    ftio_set_corrupt(&ftio, flows_corrupt);
    ftio_set_lost(&ftio, flows_lost);
    ftio_set_reset(&ftio, flows_reset);

    if (ftio_write_header(&ftio) < 0)
      fterr_errx(1, "ftio_write_header(): failed");

  }

  /* close stream */
  if (ftio_close(&ftio) < 0)
    fterr_errx(1, "ftio_close(): failed");

  /* close input */
  close (ftx.fd);

  return 0;

} /* main */

void usage() {

  fprintf(stderr, "flow-receive:\n\n");
  fprintf(stderr, " -A # replace exports with AS 0 with #\n");
  fprintf(stderr, " -b s set output byte order (little or big)\n");
  fprintf(stderr, " -C s set comment field in output\n");
  fprintf(stderr, " -d # set debug level.\n");
  fprintf(stderr, " -I ip IP address of device to capture from\n");
  fprintf(stderr, " -m m  privacy mask, ie 255.255.255.0\n");
  fprintf(stderr, " -o s  output to file s instead of stdout\n");
  fprintf(stderr, " -p # udp port to receive on.\n");
  fprintf(stderr, " -V #  version of export to capture.  Use #.# for V8\n");
  fprintf(stderr, "   1   - no sequence #, engine id, or AS\n");
  fprintf(stderr, "   5   - 1+ sequence #, engine id, and AS\n");
  fprintf(stderr, "   6   - 5+ encapsulation size\n");
  fprintf(stderr, "   7   - 5+ router shortcut (Catalyst switches)\n");
  fprintf(stderr, "   8.1 - AS aggregation\n");
  fprintf(stderr, "   8.2 - Protocol Port aggregation\n");
  fprintf(stderr, "   8.3 - Source Prefix aggregation\n");
  fprintf(stderr, "   8.4 - Destination Prefix aggregation\n");
  fprintf(stderr, "   8.5 - Prefix aggregation\n");
  fprintf(stderr, "\n\n");

} /* usage */

void sig_quit(sig)
int sig;
{
  extern int done;
  done = 1;
} /* sig_quit */

